# Code for paper: A Generative Approach for Neural Architecture Search

## Requirements
PyTorch 1.3

PyTorch Geometric 1.3.2

tqdm

sklearn

## How to run

#### Best-acc search on NAS-Bench-101
To run the two setups proposed in the paper, each for 10 runs:

```CUDA_VISIBLE_DEVICES=0 bash run_nb101_ga_nas_acc_searches.sh```

Or, to just run once:

```CUDA_VISIBLE_DEVICES=0 python3 -u run_nb101_ga_nas_acc_search.py``` 

To see the list of hyper-parameters, check ```run_nb101_ga_nas_acc_search.py```.

The results will be recorded in log files (one file for each run) inside the ```logs\``` folder. 
Scroll to the end of a log file to see the best arch found.


#### Constrained best-acc search on NAS-Bench-101
To run with the two hand-crafted cells proposed in the paper, each for 10 runs:

```CUDA_VISIBLE_DEVICES=0 bash run_nb101_ga_nas_cons_acc_searches.sh```

Or, to just run once:

```CUDA_VISIBLE_DEVICES=0 python3 -u run_nb101_ga_nas_acc_search_cons.py -constrain_source res_net``` 

Specify constrain_source as either ```res_net``` or ```inception```.

To see the list of hyper-parameters, check ```run_nb101_ga_nas_acc_search_cons.py```.

The results will be recorded in log files (one file for each run) inside the ```logs\``` folder. 
Scroll to the end of a log file to see the best arch found.


#### Best-acc search with supernet on NAS-Bench-101
Step 1 is to train a supernet with random sampling, to do that:

```CUDA_VISIBLE_DEVICES=0 python3 -u run_nb101_supernet_train.py```

Wait for the training to finish, look inside ```run_nb101_supernet_train.py``` for the list of hyper-parameters.

Step 2, to run with the trained supernet for 10 different runs:

```CUDA_VISIBLE_DEVICES=0 bash run_nb101_ga_nas_ws_acc_searches.sh```

Or, to just run once:

```CUDA_VISIBLE_DEVICES=0 python3 -u run_nb101_ga_nas_acc_search_ws.py``` 

To see the list of hyper-parameters, check ```run_nb101_ga_nas_acc_search_ws.py```.

The results will be recorded in log files (one file for each run) inside the ```logs\``` folder. 
Scroll to the end of a log file to see the best arch found.


#### Pareto front search on NAS-Bench-101
Step 1, run:

```CUDA_VISIBLE_DEVICES=0 python3 -u run_nb101_ga_nas_pareto_search.py```

Step 2, After search is done, verify that there is a file with the suffix ```..._pareto_front_values.txt``` in the ```logs/``` folder.

Step 3, run ```logs/post_process.py``` to see the Pareto front. 


#### Best-acc search on NAS-Bench-201
To run with the CIFAR-10 labelled data for 20 runs:

```CUDA_VISIBLE_DEVICES=0 bash run_nb201_ga_nas_acc_searches_c10.sh```

To run with the CIFAR-100 labelled data for 20 runs:

```CUDA_VISIBLE_DEVICES=0 bash run_nb201_ga_nas_acc_searches_c100.sh``` 

To run with the ImageNet-16-120 labelled data for 20 runs:

```CUDA_VISIBLE_DEVICES=0 bash run_nb201_ga_nas_acc_searches_img_net.sh``` 

The results will be recorded in log files (one file for each run) inside the ```logs\``` folder. 
Scroll to the end of a log file to see the best arch found.

To do a single run or to see the list of hyper-parameters, check ```run_nb201_ga_nas_acc_search.py```.


#### Best-acc search on NAS-Bench-301
First download v0.9 model files for NB301 from: https://figshare.com/articles/software/nasbench301_models_v0_9_zip/12962432

Unzip and put the ```nb_models/``` folder inside the ```libs/nasbench301-v09/``` folder

Next execute
```CUDA_VISIBLE_DEVICES=0 python3 -u run_nb301_ga_nas_cons_acc_search.py```
to test GA-NAS on NB301.

#### Constrained macro search for EfficientNet-B0
First make sure that a subsampled ImageNet-224-120 dataset is placed in the ```/data``` folder.

Next train a weight-sharing supernet for EfficientNet by: ```CUDA_VISIBLE_DEVICES=0 python3 -u run_eb0_supernet_train.py -allow_parallel```
This may take a while to finish, using multiple GPUs is strongly recommended.

Finally, execute ```CUDA_VISIBLE_DEVICES=0 python3 -u run_eb0_ga_nas_sp_cons_acc_search_ws.py ``` to start the search for EfficientNet networks.
Model size constraints can be changed through the ```-max_n_params``` flag.
