import itertools
import numpy as np
import matplotlib.cm as cm
from matplotlib import pyplot as plt
from utils.data_utils import read_col_word_delim_data


def plot_2d_evolving_pareto_front(data_file, title="Pareto front", xy_labels=("Latency", "Accuracy"),
                                  val_delim=",", col_delim="|",
                                  xy_data_indices=(2, 1), # index of the x,y values in a row of the data list
                                  length_limit=None, comment_symbols=("#", "/"),
                                  scatter_points_file=None):
    data = read_col_word_delim_data(data_file, word_delim=val_delim, col_delim=col_delim)
    markers = itertools.cycle(('P', 'o', 's', "X", "*", "D"))
    iter_values = []
    min_x, max_x = 1, -1
    for i, inst in enumerate(data):
        tag = inst[0][0]
        if tag[0] in comment_symbols: continue
        if length_limit is not None and i >= length_limit: break
        x_values = [float(v) for v in inst[xy_data_indices[0]]]
        y_values = [float(v) for v in inst[xy_data_indices[1]]]
        _v = list(zip(x_values, y_values))
        _v.sort(key=lambda t: t[0])
        x_values = [t[0] for t in _v]
        min_x = min(min_x, min(x_values))
        max_x = max(max_x, max(x_values))
        y_values = [t[1] for t in _v]
        iter_values.append((tag, x_values, y_values))
    colors = cm.rainbow(np.linspace(0, 1, len(iter_values)))
    # colors = ["blue", "red", "green", "yellow", "pink"]
    scatter_points = []
    if scatter_points_file is not None:
        with open(scatter_points_file, "r", encoding="utf-8") as f:
            lines = f.readlines()
        scatter_data = []
        for line in lines:
            line = line.rstrip()
            if len(line) == 0: continue
            if line.startswith("#") or line.startswith("/"): continue
            i, data = line.split("$")
            instances = data.split(";")
            for d_str in instances:
                _, acc, lat = d_str.split("|")
                scatter_data.append((acc, lat))
        for i, (acc, lat) in enumerate(scatter_data):
            y = float(acc)
            x = float(lat)
            scatter_points.append([x, y])
    plt.rcParams.update({'font.size': 12})
    plt.title(title)
    plt.xlabel(xy_labels[0])
    plt.ylabel(xy_labels[1])
    if len(scatter_points) > 0:
        scatter_x = [t[0] for t in scatter_points]
        scatter_y = [t[1] for t in scatter_points]
        plt.scatter(scatter_x, scatter_y, c="gray", marker=".", alpha=0.3)
    for i, (tag, x_vals, y_vals) in enumerate(iter_values):
        plt.plot(x_vals, y_vals, "{}:".format(next(markers)), color=colors[i],
                 label="{}".format(tag), linestyle="dashed")
    # max_ceil_acc = math.ceil(max([max(t[2])*100 for t in iter_values]) + 1) / 100
    # min_floor_acc = math.ceil(min([min(t[2])*100 for t in iter_values])) / 100
    # plt.yticks(np.arange(min_floor_acc, max_ceil_acc, step=0.02))
    # plt.xticks(np.arange(0.1, 1.2, step=0.05))
    # plt.grid(True, axis="both", linestyle="dashed")
    plt.legend(loc='lower right')
    plt.show()
    plt.close()


if __name__ == "__main__":
    plot_2d_evolving_pareto_front("nb101_ga_nas_pareto_pareto_front_values.txt",
                                  title="nb101_ga_nas")
    print("done")

