import utils.model_utils as m_util
from model_ppo_gan import BlockTopKAccSet
from model_ppo_gan_single_path import *
from params import *
from model_helpers import BookKeeper, ExponentialDecayOptimizer, LabelSmoothingCrossEntropy
from utils.model_utils import set_random_seed, add_weight_decay
import utils.debug_utils as d_util
from utils.data_utils import get_imagenet_train_dev_test_set, NUM_TRAIN_IMAGES_IMAGENET_120, NUM_TRAIN_IMAGES_IMAGENET

from model_ppo_gan_single_path import SPBlockState
from constants import *
from ppo_gan_pn_model import get_mask_layers_with_shortcut, make_proxylessnas_supernet, ProxylessNASSupernetEvaluator
import random
import pickle
import torch

# set the seed for reproducibility
set_random_seed(0)

mask_layers_with_shortcut = get_mask_layers_with_shortcut()
conv_candidates = list(PN_SEARCHABLE_OPS[:-1])
conv_candidates_with_shortcut = list(PN_SEARCHABLE_OPS)
conv_candidates_indices = [i for i in range(len(conv_candidates))]
conv_candidates_with_shortcut_indices = [i for i in range(len(conv_candidates_with_shortcut))] # include the Zero layer candidate
node_label2feat = {_k: torch.LongTensor([_v]) for _k, _v in PN_OP2IDX.items()}

warmup_checkpoint_path = "proxyless_nas/checkpoint/batch_256/warmup_epoch_160.pth.tar"
supernet, arch_search_run_manager = make_proxylessnas_supernet(conv_candidates, warmup_checkpoint_path)
evaluator = ProxylessNASSupernetEvaluator(supernet, arch_search_run_manager, None, None, None, None, None)

list_of_idxs = []
for i in range(0, 1000):
    # random choice the appropriate set of indices, based on if the layer has shortcut
    current_idxs = [random.choice(conv_candidates_with_shortcut_indices) if with_shortcut else random.choice(conv_candidates_indices) for with_shortcut in mask_layers_with_shortcut]
    trial = 0
    while current_idxs in list_of_idxs and trial < 10000:
        current_idxs = [random.choice(conv_candidates_with_shortcut_indices) if with_shortcut else random.choice(conv_candidates_indices) for with_shortcut in mask_layers_with_shortcut]
        trial +=1
    if current_idxs not in list_of_idxs:
        list_of_idxs.append(current_idxs)

pn_best_block = SPBlockState(PN_ORIGINAL_OPS, torch.cat([node_label2feat[lab] for lab in PN_ORIGINAL_OPS]))
pn_best_block_name = pn_best_block.get_str_id()

rand_blocks = [(pn_best_block, pn_best_block_name)]
for idxs in list_of_idxs:
    # supernet.set_chosen_op_active_with_idxs(idxs)
    # block = supernet.copy_to_normal_net()
    labels = [PN_IDX2OP_SUPERNET[idx] for idx in idxs]
    labels = [PN_INPUT_NODE_LABEL] + labels + [PN_OUTPUT_NODE_LABEL]
    block = SPBlockState(labels, torch.cat([node_label2feat[lab] for lab in labels]))
    block_name = block.get_str_id()
    rand_blocks.append((block, block_name))

counter = 0
block2acc = {}
for block, block_name in tqdm(rand_blocks, desc="Getting performance for init blocks", ascii=True):
    if block_name in block2acc:
        continue
    acc, _ = evaluator.get_perf_values(block)
    block2acc[block_name] = (block, acc)
    print("accuracy of %s : %s"%(block_name, acc))
    counter +=1
    if ((counter % 100) == 0):
        with open('checkpoint/init_1000_block2acc_dict_warmup_160.pkl', 'wb') as f:   
            pickle.dump(block2acc, f)

        with open('checkpoint/init_1000_block2acc_dict_warmup_160.txt', 'w') as f:
            for k, v in block2acc.items():
                print("%s: %s"%(k, float(v[1])), file=f)

with open('checkpoint/init_1000_block2acc_dict_warmup_160.pkl', 'wb') as f:   
    pickle.dump(block2acc, f)

with open('checkpoint/init_1000_block2acc_dict_warmup_160.txt', 'w') as f:
    for k, v in block2acc.items():
        print("%s: %s"%(k, float(v[1])), file=f)