'''
Based on run_eb0_ppo_gan_sp_cons_acc_search_ws.py
Integrate with ProxylessNAS Search Space, use ProxylessNAS supernet as evaluator
'''
import utils.model_utils as m_util
from model_ppo_gan import BlockTopKAccSet
from model_ppo_gan_single_path import *
from params import *
from model_helpers import BookKeeper, ExponentialDecayOptimizer, LabelSmoothingCrossEntropy
from utils.model_utils import set_random_seed, add_weight_decay
import utils.debug_utils as d_util
from utils.data_utils import get_imagenet_train_dev_test_set, NUM_TRAIN_IMAGES_IMAGENET_120, NUM_TRAIN_IMAGES_IMAGENET

from ppo_gan_pn_model import make_proxylessnas_supernet, ProxylessNASSupernetEvaluator
import itertools
import pickle
from os import path

def prepare_local_params(parser):
    parser.add_argument("-model_name", required=False, type=str,
                        default="eb0_ppo_gan_sp_macro_cons_acc_ws")
    parser.add_argument("-imagenet_data_dir", required=False, type=str,
                        # default=P_SEP.join([DATA_DIR, "ImageNet-224-120"]))
                        default='datasets/imagenet/')
    parser.add_argument("-supernet_checkpoint", required=False, type=str,
                        # default=P_SEP.join([SAVED_MODELS_DIR, "eb0_supernet_e500_latest.pt"]))
                        default="proxyless_nas/checkpoint/batch_256/warmup_epoch_160.pth.tar")
    parser.add_argument("-max_n_params", required=False, type=float,
                        default=5288548)
    parser.add_argument("-n_ops", required=False, type=int,
                        default=7)
    parser.add_argument("-initial_lr", required=False, type=float,
                        default=0.0001)
    parser.add_argument("-warmup_epochs", required=False, type=int,
                        default=10)
    parser.add_argument("-num_train_images", help="", type=int,
                        # default=NUM_TRAIN_IMAGES_IMAGENET_120, required=False)
                        default=NUM_TRAIN_IMAGES_IMAGENET, required=False)
    parser.add_argument("-warmup_batch_size", help="", type=int,
                        default=128, required=False)
    parser.add_argument("-warmup_base_lr", required=False, type=float,
                        default=0.016)
    parser.add_argument("-warmup_lr_decay_epoch_interval", required=False, type=float,
                        default=2.4)
    parser.add_argument("-warmup_lr_decay_ratio", required=False, type=float,
                        default=0.97)
    parser.add_argument("-top_k", required=False, type=int,
                        default=50)
    parser.add_argument("-ppo_random_init_set_size", required=False, type=int,
                        default=100)
    parser.add_argument("-C_block", required=False, type=int,
                        default=64)
    parser.add_argument("-ppo_num_iterations", required=False, type=int,
                        default=5)
    parser.add_argument("-actor_update_episode_interval", required=False, type=int,
                        default=500)
    parser.add_argument("-ppo_n_train_rounds", required=False, type=int,
                        default=10)
    parser.add_argument("-ppo_k_epochs", required=False, type=int,
                        default=10)
    parser.add_argument("-ppo_num_desired_blocks_eval", required=False, type=int,
                        default=400)
    parser.add_argument("-ppo_num_desired_blocks_inc", required=False, type=int,
                        default=100)
    parser.add_argument("-ppo_num_block_samples_eval", required=False, type=int,
                        default=None)
    parser.add_argument("-ppo_num_parallel_actors", required=False, type=int,
                        default=4)
    parser.add_argument("-ppo_parallel_explore", action="store_true", required=False,
                        default=False)
    parser.add_argument("-ppo_gan_m_repeats", required=False, type=int,
                        default=1)
    parser.add_argument("-ppo_t_timesteps", required=False, type=int,
                        default=21)
    parser.add_argument("-ppo_max_generate_steps", required=False, type=int,
                        default=21)
    parser.add_argument("-ppo_batch_size", required=False, type=int,
                        default=4096)
    parser.add_argument("-n_random_eval", help="", type=int,
                        default=0, required=False)
    parser.add_argument("-eval_only", action="store_true", required=False,
                        default=False)
    parser.add_argument("-random_mutated_truth", action="store_true", required=False,
                        default=False)
    parser.add_argument("-n_truth_per_src", help="", type=int,
                        default=100, required=False)
    parser.add_argument("-n_fake_total", help="", type=int,
                        default=500, required=False)
    parser.add_argument("-aggr_fake", action="store_true", required=False,
                        default=False)
    parser.add_argument("-ppo_reward_discount_factor", required=False, type=float,
                        default=1.0)
    parser.add_argument("-ppo_gae_lambda", required=False, type=float,
                        default=1.0)
    parser.add_argument("-ppo_clip_epsilon", required=False, type=float,
                        default=0.2)
    parser.add_argument("-ppo_c1", help="value loss", required=False, type=float,
                        default=0.5)
    parser.add_argument("-ppo_c2", help="node entropy", required=False, type=float,
                        default=0.1)
    parser.add_argument("-ppo_c3", help="edge entropy", required=False, type=float,
                        default=0.1)
    parser.add_argument("-ppo_pos_step_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_step_reward", required=False, type=float,
                        default=-0.1)
    parser.add_argument("-ppo_pos_validity_reward", required=False, type=float,
                        default=0.0)
    parser.add_argument("-ppo_neg_validity_reward",required=False,  type=float,
                        default=-0.1)
    parser.add_argument("-dis_in_channels", required=False, type=int,
                        default=128)
    parser.add_argument("-dis_hidden_size", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_out_channels", required=False, type=int,
                        default=256)
    parser.add_argument("-dis_num_layers", required=False, type=int,
                        default=2)
    parser.add_argument("-dis_dropout_prob", required=False, type=float,
                        default=0.0)
    parser.add_argument("-dis_train_batch_size", required=False, type=int,
                        default=512)
    parser.add_argument("-dis_train_epochs_per_iteration", required=False, type=int,
                        default=50)
    parser.add_argument("-dis_train_init_lr", required=False, type=float,
                        default=0.001)
    parser.add_argument("-gpn_in_channels", help="", type=int,
                        default=128, required=False)
    parser.add_argument("-gpn_hidden_size", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_out_channels", help="", type=int,
                        default=256, required=False)
    parser.add_argument("-gpn_num_layers", help="", type=int,
                        default=2, required=False)
    parser.add_argument("-gpn_dropout_prob", help="", type=float,
                        default=0.0, required=False)
    parser.add_argument("-rmsprop_eps", required=False, type=float,
                        default=0.1)
    ### new added arguments for ProxylessNAS
    parser.add_argument("-path_to_block2acc", required=False, type=str,
                        default="")
    parser.add_argument("-path_to_perf_memo", required=False, type=str,
                        default="checkpoint/perf_memo.pkl")  
    args, unknown =  parser.parse_known_args()
    return args


def dummy_lat_predictor(block, n_classes=1000):
    # as long as it smaller than params.max_n_params, there is no latency constrait
    return 6


def main(params):
    d_util.IntervalInfo.env = params.model_name
    book_keeper = BookKeeper(log_file_name=params.model_name + ".txt",
                             model_name=params.model_name, saved_model_dir=params.saved_models_dir,
                             init_eval_perf=float("-inf"), eval_perf_comp_func=lambda old, new: new > old,
                             saved_model_file=params.saved_model_file,
                             log_dir=params.logs_dir, plot_dir=params.plots_dir,
                             block_visual_dir=params.blocks_dir)
    book_keeper.log("Params: {}".format(params), verbose=False)
    set_random_seed(params.seed, log_f=book_keeper.log)

    classes = 1000
    book_keeper.log("Number of classes: {}".format(classes))
    book_keeper.log("Cand ops: {}".format(PN_SEARCHABLE_OPS[:-1]))

    supernet, arch_search_run_manager = make_proxylessnas_supernet(list(PN_SEARCHABLE_OPS[:-1]), params.supernet_checkpoint)
    n_params = sum(p.numel() for p in supernet.parameters() if p.requires_grad)
    book_keeper.log("Number of supernet trainable parameters: {}".format(n_params))
    book_keeper.log("Number of supernet total parameters: {}".format(sum(p.numel() for p in supernet.parameters())))

    evaluator = ProxylessNASSupernetEvaluator(supernet, arch_search_run_manager, None, None, None, None, None)
    if params.path_to_perf_memo != '':
        if path.exists(params.path_to_perf_memo):
            evaluator.load_perf_memo(params.path_to_perf_memo)

    node_label2feat = {_k: torch.LongTensor([_v]) for _k, _v in PN_OP2IDX.items()}
    book_keeper.log("Node features: {}".format(node_label2feat))
    book_keeper.log("Max allowable n params: {}".format(params.max_n_params))
    block_sampler = SPBlockStateSampler(PN_SEARCHABLE_OPS, n_nodes=2 + params.ppo_max_generate_steps,
                                        node_label2feat=node_label2feat)

    if params.path_to_block2acc == "":
        # Init set consists of random networks and networks with one block stacked multiple times
        rand_blocks = get_random_sp_blocks(block_sampler, params.ppo_random_init_set_size - 1)
        # Add PN best efficient GPU architecture into starting set
        pn_block = SPBlockState(PN_ORIGINAL_OPS, torch.cat([node_label2feat[lab] for lab in PN_ORIGINAL_OPS]))
        rand_blocks.append(pn_block)
    else:
        rand_blocks = None
        # will be loading the block2acc from pkl, no need for rand_blocks

    # TODO: transform the below code to ProxylessNAS space
    # Make block generator
    gpn = make_sp_gpn(in_n_nodes=len(PN_IDX2OP), out_n_nodes=len(PN_SEARCHABLE_OPS),
                      gpn_in_channels=params.gpn_in_channels, gpn_hidden_size=params.gpn_hidden_size,
                      gpn_out_channels=params.gpn_out_channels)

    # Wrap gpn into actor_critic
    actor_critic = make_sp_actor_critic(policy_net=gpn, gpn_out_channels=params.gpn_out_channels)

    # Make discriminator
    discriminator = make_sp_block_discriminator(num_unique_nodes=len(PN_IDX2OP),
                                                dis_in_channels=params.dis_in_channels,
                                                dis_hidden_size=params.dis_hidden_size)

    # Make step and final reward modules
    dis_trainer = SPDisTrainer(params.dis_train_epochs_per_iteration, discriminator,
                               params.dis_train_init_lr, params.dis_train_batch_size)
    reward_module = SPDisRewardModule(dis_trainer, evaluator, PN_OUTPUT_NODE_LABEL,
                                      params.n_random_eval, node_label2feat, PN_SEARCHABLE_OPS,
                                      n_truth_per_src=params.n_truth_per_src, n_fake_total=params.n_fake_total,
                                      enable_mutation=params.random_mutated_truth, aggr_fake=params.aggr_fake,
                                      pos_reward=params.ppo_pos_validity_reward,
                                      neg_reward=params.ppo_neg_validity_reward)

    # Define initial state
    init_state = get_sp_init_state(PN_INPUT_NODE_LABEL, node_label2feat)

    # Make environments
    env_manager = make_sp_env_manager(actor_critic=actor_critic, n_parallel_actors=params.ppo_num_parallel_actors,
                                      init_state=init_state, node_label2feat=node_label2feat,
                                      reward_module=reward_module,
                                      max_allowed_steps=params.ppo_max_generate_steps,
                                      output_node_label=PN_OUTPUT_NODE_LABEL,
                                      scaffold_node_labels=PN_SEARCHABLE_OPS)

    actor_critic = env_manager.get_actor_rep()
    value_criterion = torch.nn.MSELoss()
    actor_optimizer = torch.optim.Adam(actor_critic.parameters(), lr=params.initial_lr)
    book_keeper.log(actor_critic)
    book_keeper.log("Actor critic for {}".format(params.model_name))
    n_params = sum(p.numel() for p in actor_critic.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))
    book_keeper.log(discriminator)
    book_keeper.log("Discriminator for {}".format(params.model_name))
    n_params = sum(p.numel() for p in discriminator.parameters() if p.requires_grad)
    book_keeper.log("Number of trainable parameters: {}".format(n_params))

    if params.path_to_block2acc != "":
        with open(params.path_to_block2acc, 'rb') as f:   
            block2acc = pickle.load(f)
        block2acc = dict(itertools.islice(block2acc.items(), 0, params.ppo_random_init_set_size))
    else:
        counter = 0
        block2acc = {}
        for block in tqdm(rand_blocks, desc="Getting performance for init blocks", ascii=True):
            if block.get_str_id() in block2acc:
                continue
            if not params.eval_only:
                acc, _ = evaluator.get_perf_values(block)
            else:
                acc = random.random()
            block2acc[block.get_str_id()] = (block, acc)
            counter +=1
            if ((counter % 100) == 0):
                with open('checkpoint/init_%s_block2acc_dict_warmup_160.pkl'%(params.ppo_random_init_set_size), 'wb') as f:   
                    pickle.dump(block2acc, f)
                with open('checkpoint/init_%s_block2acc_dict_warmup_160.txt'%(params.ppo_random_init_set_size), 'w') as f:
                    for k, v in block2acc.items():
                        print("%s: %s"%(k, float(v[1])), file=f)
        with open('checkpoint/init_%s_block2acc_dict_warmup_160.pkl'%(params.ppo_random_init_set_size), 'wb') as f:   
            pickle.dump(block2acc, f)
        with open('checkpoint/init_%s_block2acc_dict_warmup_160.txt'%(params.ppo_random_init_set_size), 'w') as f:
            for k, v in block2acc.items():
                print("%s: %s"%(k, float(v[1])), file=f)

    book_keeper.log("Initializing Top-{} acc set".format(params.top_k))
    book_keeper.log("Initial block2acc size: {}".format(len(block2acc)))
    top_acc_set = BlockTopKAccSet(block2acc, params.top_k, log_f=book_keeper.log)
    book_keeper.log("Initial top acc set size: {}".format(len(top_acc_set.top_block2acc())))

    if not params.eval_only:
        book_keeper.log("Search")
        # Can work under the PT search routine
        sp_ppo_gan_cons_pt_acc_search(env_manager=env_manager, ppo_optimizer=actor_optimizer,
                                      value_criterion=value_criterion, top_acc_set=top_acc_set,
                                      evaluator=evaluator, n_train_iterations=params.ppo_num_iterations,
                                      n_train_rounds=params.ppo_n_train_rounds,
                                      k_epochs=params.ppo_k_epochs, t_timesteps=params.ppo_t_timesteps,
                                      batch_size=params.ppo_batch_size,
                                      n_desired_blocks_eval=params.ppo_num_desired_blocks_eval,
                                      n_desired_blocks_inc=params.ppo_num_desired_blocks_inc,
                                      output_node_label=PN_OUTPUT_NODE_LABEL,
                                      max_lat=params.max_n_params, ext_lat_predictor=dummy_lat_predictor,
                                      book_keeper=book_keeper,
                                      parallel_explore=params.ppo_parallel_explore,
                                      m_repeats=params.ppo_gan_m_repeats,
                                      reward_discount_factor=params.ppo_reward_discount_factor,
                                      gae_lambda=params.ppo_gae_lambda, clip_epsilon=params.ppo_clip_epsilon,
                                      c1=params.ppo_c1, c2=params.ppo_c2,
                                      actor_update_episode_interval=params.actor_update_episode_interval)
    else:
        book_keeper.load_state_dict_checkpoint(top_acc_set, "{}_top_acc_set.pkl".format(params.model_name),
                                               allow_silent_fail=False)
        book_keeper.log("Loaded top acc set size: {}".format(len(top_acc_set)))

    book_keeper.log("Evaluation")
    new_blocks = top_acc_set.get_top_blocks_with_acc()
    book_keeper.log("{} new nets proposed for evaluation".format(len(new_blocks)))
    for i, (b, acc) in enumerate(new_blocks):
        book_keeper.log("Net {}: {}".format(i, b))
        book_keeper.log("Searched acc: {}".format(acc))
        book_keeper.log("N params: {}".format(dummy_lat_predictor(b)))

    if params.path_to_perf_memo != '':
        evaluator.save_perf_memo(params.path_to_perf_memo)

if __name__ == "__main__":
    _parser = prepare_global_params()
    _args = prepare_local_params(_parser)
    m_util.DEVICE_STR_OVERRIDE = _args.device_str
    main(_args)
    print("done")
