class Params(dict):
    def __init__(self, *arg, **kw):
        super(Params, self).__init__(*arg, **kw)
        self.__dict__ = self


class ParamsIterator:

    def __init__(self, *args):
        self.param_lists = [a for a in args]
        if not self.param_lists:
            raise RuntimeError("Parameters cannot be empty")
        self.param_list_lens = [len(x)-1 for x in self.param_lists]
        self.curr_iters = [0 for _ in range(len(self.param_lists))]
        self.curr_iter_idx = 0

    def _update_iter(self):
        if self.iteration_done():
            return
        if self.curr_iters[self.curr_iter_idx] == self.param_list_lens[self.curr_iter_idx]:
            while self.curr_iter_idx < len(self.param_lists) \
                and self.curr_iters[self.curr_iter_idx] == self.param_list_lens[self.curr_iter_idx]:
                self.curr_iter_idx += 1
            if self.curr_iter_idx >= len(self.param_lists):
                assert(self.iteration_done())
            else:
                self.curr_iters[self.curr_iter_idx] += 1
                for j in range(self.curr_iter_idx):
                    self.curr_iters[j] = 0
                self.curr_iter_idx = 0
        else:
            self.curr_iters[self.curr_iter_idx] += 1

    def iter_started(self):
        return self.curr_iter_idx == 0

    def get_next_combinations(self):
        rv = []
        for i in range(len(self.param_lists)):
            val = self.param_lists[i][self.curr_iters[i]]
            rv.append(val)
        self._update_iter()
        return rv

    def iteration_done(self):
        return self.curr_iter_idx >= len(self.param_lists)

    def reset(self):
        self.curr_iters = [0 for _ in range(len(self.param_lists))]
        self.curr_iter_idx = 0


class WrapperOptimizer:
    # TODO
    def __init__(self, name, optimizer,
                 step_routine,
                 pre_step_routines=[],
                 post_step_routines=[],
                 performances=[]):
        self.name = name
        self._optimizer = optimizer
        self.step = step_routine
        self.pre_step_routines = pre_step_routines
        self.post_step_routines = post_step_routines
        self.performances = performances

    def reset_performance_record(self):
        self.performances = []


class OptimizerSelector:
    # TODO
    def __init__(self, named_optim_dict, n_trails=5, report_filename="optim_report.txt"):
        self.n_trails = n_trails
        self._named_optim_dict = named_optim_dict
        self._named_optim_list = [(n, v) for n, v in named_optim_dict.items()]
        self._report_filename = report_filename
        self._curr_idx = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def __len__(self):
        return len(self._named_optim_list)

    def has_next(self):
        return self._curr_idx < len(self._named_optim_list)

    def reset(self):
        self._curr_idx = 0
        for _, optim in self._named_optim_list:
            optim.reset_performance_record()

    def next(self):
        if self._curr_idx >= len(self._named_optim_list):
            self.reset()
            raise StopIteration()
        name, optim = self._named_optim_list[self._curr_idx]
        self._curr_idx += 1
        return name, optim


if __name__ == "__main__":

    it = ParamsIterator([1,2,3], ['a', 'b'], [-1, -2, -3, -4])

    while not it.iteration_done():
        li = it.get_next_combinations()
        print(li)

    print("")

    it = ParamsIterator([1, 2], ['a'])

    while not it.iteration_done():
        li = it.get_next_combinations()
        print(li)

